
var exports = module.exports = require('connect').middleware;
var Server = exports.Server = require('express/server');
var Step = require('step');

var fs = require('fs'),
    express = require('express'),
    underscore = require('underscore'),
    couchdb = require('couchdb'),
    utils = require('./utils'),
    middleware = require('./middleware'),
    RedisStore = require('connect-redis'),
    Storage = require('./storage').Storage,
    opts = {},
    settings = null;

exports.Storage = Storage;

exports.createServer = function() {
  return new Server(Array.prototype.slice.call(arguments));
};

exports.go = function(app, options, fn) {
    console.info('Booting M1 Node');
    _.extend(opts, options);
    settings = require(opts.homeDir + '/settings');
    app.stores = new Storage(settings);
    Step(
        function() {
            bootCouchViews(app, this);
        },
        function(err, val) {
            if (err) throw err;
            bootApplication(app, this);
        },
        function(err, val) {
            if (err) throw err;
            bootSubApps(app, this);
        },
        function(err, val) {
            if (err) throw err;
            bootRoutes(app)
            return true;
        },
        function(err, val) {
            if (err) throw err;
            fn();
        }
    );
};

function bootSubApps(app, fn) {
    console.info('Booting sub apps');
    try {
        boot = require(opts.homeDir + '/boot');
        boot.go(app, fn);
    } catch (err) {
        // it's probably fine when this errors but it could be good to know about
        console.log('Error booting sub-application')
        console.dir(err)
        fn();
    }
}

function bootApplication(app, fn) {
    console.info('Booting application');
    app.use(express.logger({ format: ':method :url :status' }));
    app.use(express.bodyDecoder());
    app.use(express.methodOverride());
    app.use(express.cookieDecoder());
    var oneHour = 60 * 60 * 1000;
    app.use(express.session({
        store: new RedisStore({
            maxAge: settings.session_timeout || oneHour
        })
    }));
    var loadProfile = require('./middleware').loadProfile;
    app.use(loadProfile);
    app.use(app.router);
    app.use(express.staticProvider(opts.homeDir + '/public'));

    app.error(function(err, req, res){
        console.error('bootApplication: app callback error');
        console.dir(err);
        res.render(opts.homeDir + '/views/500');
    });

    app.use(function(req, res){
        res.render('404');
    });

    var templateLanguage = settings.template_language || 'ejs';
    app.set('views', opts.homeDir + '/views');
    app.register('.html', require(templateLanguage));
    app.set('view engine', 'html');

    app.dynamicHelpers({
        request: function(req){
            return req;
        },
        messages: function(req, res){
            return function(){
                var messages = req.flash();
                return res.partial('messages', {
                    object: messages,
                    as: 'messageTypes',
                    locals: { hasMessages: Object.keys(messages).length },
                    dynamicHelpers: false
                });
            }
        }
    });
    fn();
}

function bootCouchViews(app, fn) {
    console.info('Booting couch views');
    var views = require(opts.homeDir + '/store/views');
    var stores = app.stores.getDatabases();
    var steps = [];

    steps.push(function start() {
        // prime the pump
        return null;
    });

    steps.push(function processViews(err, val) {
        if (err) throw err;
        var group = this.group();
        _.each(stores, function(store, name) {
            if (views[name]) {
                _.each(views[name], function(design, designName) {
                    var id = '_design/' + designName;
                    var saveDesign = {
                        id: id, 
                        views: design
                    };
                    var noChange = false;
                    var _group = group();
                    store.getDoc(id, function(err, designDoc) {
                        try {
                            if (!err && (couchdb.toJSON(designDoc.views) !== couchdb.toJSON(design))) {
                                designDoc.views = design;
                                saveDesign = designDoc;
                            } else if (!err) {
                                noChange = true;
                            }
                        } catch (err) {
                            console.dir(err);
                        }
                        if (!noChange) {
                            store.saveDesign(id, saveDesign, function(err, ok) {
                                if (err) console.dir(err);
                                ok.name = name;
                                _group(err, ok);
                            });
                        } else {
                            _group(null, null);
                        }
                    });
                });
            }
        });
    });

    steps.push(function end(err, vals) {
        if (err) throw err;
        var updates = _.compact(vals);
        if (updates.length > 0) {
            _.each(updates, function(update) {
                console.info(' * Updated ' + update.id + ' in ' + update.name + ' database');
            });
        } else {
            console.log(' * All views are up to date');
        }
        fn();
    });

    Step.apply(this, steps);
}

function bootRoutes(app) {
    console.info('Booting routes');
    var urls = require(opts.homeDir + '/urls').patterns();
    _.each(urls, function(route) {
        bootRoute(app, route);
    });
}

function parseRoute(rawRoute) {
    var pattern = rawRoute[0];
    var rawAction = rawRoute[1];
    var middleware = rawRoute[2];
    var parts = rawAction.split('.');
    var actionName = parts[(parts.length-1)];
    var controllerParts = parts.slice(0, (parts.length-1));
    var controllerPath = controllerParts.join('/');
    var controller = require(opts.homeDir + '/controllers/' + controllerPath);
    var action = controller[actionName];
    return {
        pattern: pattern,
        action: action,
        controllerPath: controllerPath,
        actionName: actionName,
        middleware: middleware
    };
}

function registerRoute(app, route, fn) {
    if (_.isUndefined(route.middleware)) {
        app.get(route.pattern, fn);
        app.post(route.pattern, fn);
    } else {
        app.get(route.pattern, route.middleware, fn);
        app.post(route.pattern, route.middleware, fn);
    }
}

function bootRoute(app, rawRoute) {
    var route = parseRoute(rawRoute),
        fn = controllerAction(route);
    registerRoute(app, route, fn);
}

function controllerAction(route) {
    return function(req, res, next) {
        var render = res.render,
            format = req.params.format,
            path = opts.homeDir + '/views/' + route.controllerPath + '/' + route.actionName + '.html';
        res.render = function(obj, options, fn) {
            res.render = render;
            // Template path
            if (_.isString(obj)) {
                return res.render(obj, options, fn);
            }
            // Format support
            if (route.actionName == 'show' && format) {
                if (format === 'json') {
                    return res.send(obj);
                } else {
                    throw new Error('unsupported format "' + format + '"');
                }
            }
            // Render template
            res.render = render;
            options = options || {};
            options.locals = options.locals || {};
            // Expose obj as the "users" or "user" local
            // Do we want and/or need this?
            options.locals[route.controllerPath] = obj;
            return res.render(path, options, fn);
        };
        route.action.apply(this, arguments);
    };
}
